/* -*- Mode: Vala; indent-tabs-mode: nil; c-basic-offset: 2; tab-width: 2 -*- */
/*
    Multiplication Puzzle
    For copyright information, see AUTHORS.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using GLib;

public class GtkMult : Gtk.ApplicationWindow
{
  public static int dpi;
  
  public MultPuzzle puzzle { get; private set; }
  public bool cheated { get; private set; }
  
  public const double ASPECT = 8.0/6.0;
  
  public void set_feedback(string feedback, bool timeout=true)
  {
    if (status_timeout_id != 0)
      Source.remove (status_timeout_id);

    clear_feedback();
    header.title = feedback;

    if (timeout)
      status_timeout_id = Timeout.add_seconds(5, clear_feedback);
  }

  public GtkMult(Gtk.Application app)
  {
    Object(application: app);
  }
  
  construct
  {
    vbox = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);
    score = new Gtk.Label ("");
    canvas = null;
    status_timeout_id = 0;
    timer = new Timer();
    timer_timeout_id = 0;
    
    // Convert pixel space to desired point size
    int xft_dpi;
    var settings = Gtk.Settings.get_for_screen (get_screen ());
    settings.get ("gtk-xft-dpi", out xft_dpi);
    dpi = xft_dpi / Pango.SCALE;
    
    var height = 4*dpi;
    set_default_size((int)(height*ASPECT), height);
    
    Gtk.Box vb = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);
    vb.pack_start (vbox, true, true, 0);
    add (vb);
    
    add_events (Gdk.EventMask.KEY_PRESS_MASK);
    key_press_event.connect(handle_key_press);
    
    set_title (_("Multiplication Puzzle"));

    var new_button = new Gtk.Button.with_mnemonic(_("_New Game"));
    new_button.clicked.connect(new_puzzle);

    hint_button = new Gtk.Button.from_icon_name("dialog-question-symbolic");
    hint_button.set_tooltip_text(_("Reveal a digit at random"));
    hint_button.clicked.connect(show_hint);

    header = new Gtk.HeaderBar();
    header.show_close_button = true;
    header.pack_start(new_button);
    header.pack_end(hint_button);
    set_titlebar(header);

    new_puzzle ();
  }
  
  static unichar a_char = _("A").get_char().toupper();
  static unichar b_char = _("B").get_char().toupper();
  static unichar c_char = _("C").get_char().toupper();
  static unichar d_char = _("D").get_char().toupper();
  static unichar e_char = _("E").get_char().toupper();
  static unichar f_char = _("F").get_char().toupper();
  static unichar g_char = _("G").get_char().toupper();
  static unichar h_char = _("H").get_char().toupper();
  static unichar i_char = _("I").get_char().toupper();
  static unichar j_char = _("J").get_char().toupper();
  static unichar 0_char = _("0").get_char().toupper();
  static unichar 1_char = _("1").get_char().toupper();
  static unichar 2_char = _("2").get_char().toupper();
  static unichar 3_char = _("3").get_char().toupper();
  static unichar 4_char = _("4").get_char().toupper();
  static unichar 5_char = _("5").get_char().toupper();
  static unichar 6_char = _("6").get_char().toupper();
  static unichar 7_char = _("7").get_char().toupper();
  static unichar 8_char = _("8").get_char().toupper();
  static unichar 9_char = _("9").get_char().toupper();
  private bool handle_key_press (Gdk.EventKey event)
  {
    // Don't recognize any keys if ctrl or alt are pressed.
    var state = (event.state & Gtk.accelerator_get_default_mod_mask());
    state = state & (~0x03); // ignores shift and caps lock
    if (state != 0)
      return false;
    
    char ch;
    unichar ev_char = Gdk.keyval_to_unicode(Gdk.keyval_to_upper(event.keyval));
    
    if      (ev_char == a_char) ch = 'A';
    else if (ev_char == b_char) ch = 'B';
    else if (ev_char == c_char) ch = 'C';
    else if (ev_char == d_char) ch = 'D';
    else if (ev_char == e_char) ch = 'E';
    else if (ev_char == f_char) ch = 'F';
    else if (ev_char == g_char) ch = 'G';
    else if (ev_char == h_char) ch = 'H';
    else if (ev_char == i_char) ch = 'I';
    else if (ev_char == j_char) ch = 'J';
    else if (ev_char == 0_char) ch = '0';
    else if (ev_char == 1_char) ch = '1';
    else if (ev_char == 2_char) ch = '2';
    else if (ev_char == 3_char) ch = '3';
    else if (ev_char == 4_char) ch = '4';
    else if (ev_char == 5_char) ch = '5';
    else if (ev_char == 6_char) ch = '6';
    else if (ev_char == 7_char) ch = '7';
    else if (ev_char == 8_char) ch = '8';
    else if (ev_char == 9_char) ch = '9';
    else if (event.keyval == 65307) { // escape; vala needs keysyms
      this.canvas.clear_mode();
      return true;
    }
    else
      return false;
    
    this.canvas.start_choice(ch);
    return true;
  }
    
  private Canvas canvas;
  private Gtk.Box vbox;
  private Timer timer;
  private Gtk.Label score;
  Gtk.HeaderBar header;
  Gtk.Button hint_button;
  private uint status_timeout_id;
  private uint timer_timeout_id;
  
  public void new_puzzle ()
  {
    if (canvas != null)
      vbox.remove (canvas);

    puzzle = new MultPuzzle (3, 2);
    
    canvas = new Canvas (this);
    vbox.pack_start (canvas, true, true, 0);
    
    Gdk.Geometry geom = Gdk.Geometry();
    geom.min_aspect = ASPECT;
    geom.max_aspect = ASPECT;
    set_geometry_hints(canvas, geom, Gdk.WindowHints.ASPECT);
    
    timer.start();
    if (timer_timeout_id != 0)
      Source.remove(timer_timeout_id);
    timer_timeout_id = Timeout.add_seconds (1, update_subtitle);
    
    cheated = false;
    
    vbox.show_all ();
    on_puzzle_change(puzzle);

    clear_feedback();
    update_subtitle();

    if (status_timeout_id != 0)
      Source.remove (status_timeout_id);
    status_timeout_id = 0;
    
    puzzle.guessed.connect(on_puzzle_guess);
    puzzle.changed.connect(on_puzzle_change);
  }
  
  private void on_puzzle_guess (MultPuzzle p, int digit, MultPuzzleChar letter, MultPuzzleGuessStatus response)
  {
    update_subtitle();

    if (puzzle.is_done)
    {
      set_feedback(_("Congratulations!"), false);
    }
    else
    {
      string message_pattern = null;
      
      switch (response)
      {
      case MultPuzzleGuessStatus.WRONG:
        // Translators: First argument is letter, second is digit
        message_pattern = _("Incorrect — %1$s is not %2$s");
        break;
      case MultPuzzleGuessStatus.CORRECT:
        // Translators: First argument is letter, second is digit
        message_pattern = _("Correct — %1$s is %2$s");
        break;
      default:
      case MultPuzzleGuessStatus.KNOWN:
      case MultPuzzleGuessStatus.INVALID:
        // shouldn't happen
        break;
      }
      
      if (message_pattern != null) {
        var letter_str = "%c".printf ((char)letter);
        var digit_str = "%i".printf (digit);
        set_feedback(message_pattern.printf(_(letter_str), _(digit_str)));
      }
    }
  }
  
  private void on_puzzle_change (MultPuzzle p)
  {
    hint_button.sensitive = !p.is_done;
    if (p.is_done)
      timer.stop();
  }

  bool clear_feedback()
  {
    status_timeout_id = 0;
    header.title = _("Multiplication Puzzle");
    return false;
  }

  bool update_subtitle()
  {
    var wrong = puzzle.wrong_guesses;
    var guesses = "%s: %i%s".printf(_("Incorrect Guesses"), wrong, cheated ? "*" : "");

    var secs = (uint)timer.elapsed(null);
    var time = secs > 60 * 60 ?
               "%02u:%02u:%02u".printf(secs/60/60, secs/60%60, secs%60) :
               "%02u:%02u".printf(secs/60%60, secs%60);

    header.subtitle = "%s   %s".printf(guesses, time);

    return true;
  }

  public void show_hint()
  {
    cheated = true;
    
    // Pick a random unknown digit.  Keep trying until we actually solve one
    // (because the digit may not have been in the puzzle -- we want to be
    // more useful than that.
    var character = MultPuzzleChar.INVALID;
    var digit = 0;
    var needed = this.puzzle.get_needed_digits();
    var num_needed = 0;
    foreach (bool n in needed)
      num_needed += n ? 1 : 0;
    if (num_needed == 0) // shouldn't happen
      return;
    
    var choice = Random.int_range(0, num_needed);
    for (int i = 0; i < 10; ++i) {
      if (needed[i] && choice == 0) {
        character = this.puzzle.solve_digit(i);
        digit = i;
        break;
      }
      else if (needed[i])
        --choice;
    }

    // Translators: First argument is letter, second is digit
    var message_pattern = _("%1$s is %2$s");
    var letter_str = "%c".printf ((char)character);
    var digit_str = "%i".printf (digit);
    set_feedback(message_pattern.printf(_(letter_str), _(digit_str)));

    update_subtitle();
  }
}

